
# This file was generated by 'versioneer.py' (0.21) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2022-12-07T15:15:50-0800",
 "dirty": false,
 "error": null,
 "full-revisionid": "3421b13dff0cc71717998b62d17e9fb6bb65f5d1",
 "version": "0.2.6"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
