from .credentials import DbtCliProfile  # noqa

from .configs.base import (  # noqa
    TargetConfigs,
    GlobalConfigs,
    MissingExtrasRequireError,
)

try:
    from .configs.snowflake import SnowflakeTargetConfigs  # noqa
except MissingExtrasRequireError:
    pass

try:
    from .configs.bigquery import BigQueryTargetConfigs  # noqa
except MissingExtrasRequireError:
    pass

try:
    from .configs.postgres import PostgresTargetConfigs  # noqa
except MissingExtrasRequireError:
    pass
